"""This module implements the mosaik simulator API for the
pandapower model.

"""

import json
import logging
import sys
from typing import Any, Dict, List

import mosaik_api_v3
import numpy as np
from midas.util.dict_util import strtobool, update
from midas.util.logging import set_and_init_logger
from midas.util.print_format import mformat
from midas.util.runtime_config import RuntimeConfig

from .meta import META
from .model.static import PandapowerGrid
from .plotter import Plotter

# Need to hard code the name because if started as separate
# procress, the __name__ is __main__ instead.
LOG = logging.getLogger("midas_powergrid.simulator")


class PandapowerSimulator(mosaik_api_v3.Simulator):
    """The pandapower simulator."""

    def __init__(self):
        super().__init__(META)
        self.sid: str = ""
        self.models: Dict[str, PandapowerGrid] = {}
        self.entity_map: dict = {}
        self.entities: list = []

        self.step_size: int = 0
        self.now_dt = None
        self.cache = dict()

        self.plotting = False
        self.plotter_cfg = dict()
        self.plotter = None
        self._sim_time: int = 0
        self.key_value_logs: bool

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.

        Returns
        -------
        dict
            The meta dict (set by mosaik_api.Simulator)

        """
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                + "Using default step_size of 900."
            )

        self.step_size = sim_params.get("step_size", 900)

        self.plotting = sim_params.get("plotting", False)
        if not isinstance(self.plotting, bool):
            try:
                self.plotting = strtobool(self.plotting)
            except AttributeError:
                self.plotting = False

        if self.plotting:
            self.plotter = Plotter()
            self.plotter_cfg["plot_path"] = sim_params.get(
                "plot_path", "_plots"
            )

        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity)

        Returns
        -------
        list
            A list with information on the created entity.

        """

        entities = []
        # use_constraints = model_params.get("use_constraints", False)

        for _ in range(num):
            gidx = len(self.models)
            eid = f"{model}-{gidx}"

            if "gridfile" not in model_params:
                LOG.debug(
                    "Param *gridfile* not provided in model_params. "
                    "Using default grid *midasmv*"
                )

            self.models[eid] = PandapowerGrid(model_params, gidx)

            self.plotter_cfg[eid] = {
                "grid_type": self.models[eid].grid_type,
                "plot_name": model_params["gridfile"],
                "plotting": self.plotting,
            }

            children = list()
            for cid, attrs in sorted(self.models[eid].entity_map.items()):
                assert cid not in self.entity_map
                self.entity_map[cid] = attrs
                etype = attrs["etype"]

                # We'll only add relations from lines to nodes (and not
                # from nodes to lines) because this is sufficient for
                # mosaik to build the entity graph
                relations = list()
                if etype.lower() in [
                    "trafo",
                    "line",
                    "load",
                    "sgen",
                    "storage",
                ]:
                    relations = attrs["related"]

                children.append({"eid": cid, "type": etype, "rel": relations})

            entity = {
                "eid": eid,
                "type": model,
                "rel": list(),
                "children": children,
            }
            entities.append(entity)
            self.entities.append(entity)

        return entities

    def step(self, time, inputs, max_advance=0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """
        self._sim_time = time

        for eid, attrs in inputs.items():
            log_msg = {
                "id": f"{self.sid}.{eid}",
                "name": eid,
                "type": eid.split("-")[1],
                "sim_time": self._sim_time,
                "msg_type": "input",
                "src_eids": [],
            }
            gidx = eid.split("-")[0]
            grid = self.models.get(f"Grid-{gidx}", None)

            if grid is None:
                LOG.critical("No grid found for grid index %s!", gidx)
                raise KeyError

            idx = self.entity_map[eid]["idx"]
            etype = self.entity_map[eid]["etype"]

            for attr, src_ids in attrs.items():
                setpoint = 0.0
                all_none = True
                for src_id, val in src_ids.items():
                    if val is not None:
                        setpoint += float(val)
                        all_none = False
                        log_msg["src_eids"].append(src_id)
                if all_none:
                    setpoint = None

                attrs[attr] = setpoint
                log_msg[attr] = setpoint
            try:
                grid.set_inputs(etype, idx, attrs)
            except Exception as err:
                LOG.error(
                    f"Failed to set inputs {attrs} for etype {etype}"
                    f" at index {idx}: {err}"
                )
                raise ValueError(
                    f"Impossible input {attrs} for {etype} ({idx})."
                )
            log_msg["src_eids"] = list(set(log_msg["src_eids"]))
            LOG.info(json.dumps(log_msg))

        for eid, model in self.models.items():
            try:
                model.run_powerflow(time)
            except Exception as err:
                LOG.error(f"Failed to perform power flow calculation: {err}")
                raise ValueError(
                    f"Failed to perform power flow calculation: {err}"
                )

            try:
                self.cache[eid] = model.get_outputs()
                for child, cache in self.cache[eid].items():
                    self.cache[child] = cache
            except Exception as err:
                LOG.error(f"Failed to get outputs: {err}")
                raise ValueError(f"Failed to get outputs: {err}")

            if self.plotter is not None and self.plotter_cfg[eid]["plotting"]:
                self.plotter.grid = model.grid._grid
                self.plotter.plot_path = self.plotter_cfg["plot_path"]
                self.plotter.plot_name = self.plotter_cfg[eid]["plot_name"]
                self.plotter.grid_type = self.plotter_cfg[eid]["grid_type"]
                self.plotter.plot(eid, int(time / self.step_size))

        return time + self.step_size

    def get_data(
        self, outputs: Dict[str, List[str]]
    ) -> Dict[str, Dict[str, Any]]:
        """Return the requested outputs (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """

        data: dict = {}
        grid_json: dict = {}

        for eid, attrs in outputs.items():
            if "Grid" in eid:
                self._get_grid_attributes(eid, attrs, data, grid_json)
            else:
                self._get_children_attributes(eid, attrs, data)

        update(data, grid_json)
        return data

    def finalize(self):
        for model in self.models.values():
            model.finalize()

    def _get_grid_attributes(
        self,
        eid: str,
        attrs: List[str],
        data: Dict[str, Dict[str, Any]],
        grid_json: Dict[str, Any],
    ) -> None:
        if "health" in attrs:
            health = (
                self.models[eid]
                .grid.get_value("res_bus", attr="vm_pu")[1:]
                .mean()
            )
            if np.isnan(health):
                LOG.debug(
                    "Grid health is NaN. This might be okay so "
                    "setting health to 0. Just that you know!"
                )
                health = 0
            data.setdefault(eid, dict())["health"] = health

        if "grid_json" in attrs:
            grid_json.setdefault(eid, dict())["grid_json"] = self.models[
                eid
            ].to_json()

    def _get_children_attributes(
        self, eid: str, attrs: List[str], data: Dict[str, Dict[str, Any]]
    ) -> None:
        log_msg = {
            "id": f"{self.sid}_{eid}",
            "name": eid,
            "type": eid.split("-")[1],
            "sim_time": self._sim_time,
            "msg_type": "output",
        }
        for attr in attrs:
            val = self.cache[eid][attr]
            if np.isnan(val):
                LOG.debug(f"{attr} for {eid} is NaN. Setting to 0.")
                val = 0
            if isinstance(val, np.bool_):
                val = bool(val)
            if isinstance(val, (np.int32, np.int64)):
                val = int(val)
            if isinstance(val, (np.float32, np.float64)):
                val = float(val)

            data.setdefault(eid, dict())[attr] = val
            log_msg[attr] = val

        try:
            LOG.info(json.dumps(log_msg))
        except TypeError:
            print(log_msg)
            sys.exit(-1)


if __name__ == "__main__":
    set_and_init_logger(
        0, "powergrid-logfile", "midas-powergrid.log", replace=True
    )
    LOG.info("Starting mosaik simulation...")
    mosaik_api_v3.start_simulation(PandapowerSimulator())



"""This module contains test cases for the pandapower grid model."""

import unittest

import numpy as np

from midas_powergrid.model.static import PandapowerGrid


class TestPandapowerGrid(unittest.TestCase):
    """Test case for the pandapower grid wrapper."""

    def test_cigre_lv(self):
        """Test for *common* pandapower grids."""
        model = PandapowerGrid({"gridfile": "cigre_lv"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_cigre_mv(self):
        """Test for *common* pandapower grids."""
        model = PandapowerGrid({"gridfile": "cigre_mv"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_cigre_hv(self):
        """Test for *common* pandapower grids."""
        model = PandapowerGrid({"gridfile": "cigre_hv"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_midas_mv(self):
        """Test for midas pandapower grid variants."""
        model = PandapowerGrid({"gridfile": "midasmv"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_midas_lv(self):
        """Test for midas pandapower grid variants."""
        model = PandapowerGrid({"gridfile": "midaslv"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_simbench(self):
        """Test for simbench grids."""
        model = PandapowerGrid({"gridfile": "1-LV-rural3--0-sw"})

        model.run_powerflow(0)
        outputs = model.get_outputs()
        self.assertTrue(outputs)

    def test_json(self):
        """Test for a json grid."""
        pass

    def test_excel(self):
        """Test for a xlsx grid."""
        pass

    def test_set_inputs_load(self):
        """Test to set an input for a load."""
        model = PandapowerGrid({"gridfile": "cigre_lv"})

        self.assertEqual(model.grid.get_value("load", 0, "p_mw"), 0.19)
        self.assertEqual(model.grid.get_value("load", 0, "q_mvar"), 0.06244998)
        self.assertTrue(model.grid.get_value("load", 0, "in_service"))

        model.set_inputs(
            etype="Load",
            idx=0,
            data={"p_mw": 0.04, "q_mvar": 0.02, "in_service": False},
        )

        self.assertEqual(model.grid.get_value("load", 0, "p_mw"), 0.04)
        self.assertEqual(model.grid.get_value("load", 0, "q_mvar"), 0.02)
        self.assertFalse(model.grid.get_value("load", 0, "in_service"))

    def test_get_outputs(self):
        """Test to get the outputs after the powerflow."""

        model = PandapowerGrid({"gridfile": "simple_four_bus_system"})
        output = model.get_outputs()

        self.assertAlmostEqual(output["0-bus-1"]["vm_pu"], 0.996608)
        self.assertAlmostEqual(
            output["0-bus-1"]["va_degree"], -150.208, places=3
        )

        self.assertAlmostEqual(
            output["0-line-0"]["loading_percent"], 31.273, places=3
        )
        self.assertAlmostEqual(
            output["0-trafo-0"]["va_lv_degree"], -150.208, places=3
        )

        self.assertEqual(output["0-load-0-2"]["p_mw"], 0.03)
        self.assertEqual(output["0-sgen-1-3"]["p_mw"], 0.015)

    def test_trafo_pos_input(self):
        model = PandapowerGrid({"gridfile": "midasmv"})
        model.set_inputs(etype="trafo", idx=0, data={"tap_pos": 0.342})
        outputs = model.get_outputs()
        self.assertEqual(3, outputs["0-trafo-0"]["tap_pos"])
        model.set_inputs(etype="trafo", idx=1, data={"tap_pos": None})
        outputs = model.get_outputs()
        self.assertEqual(0, outputs["0-trafo-1"]["tap_pos"])
        model.set_inputs(etype="trafo", idx=1, data={"tap_pos": np.NaN})
        outputs = model.get_outputs()
        self.assertEqual(0, outputs["0-trafo-1"]["tap_pos"])


if __name__ == "__main__":
    unittest.main()



"""In this module, the midas lv grid is defined and configured."""
import pandapower as pp
import pandapower.networks as pn


def build_grid(**kwargs):
    grid = pn.create_cigre_network_lv()
    grid.load = grid.load[0:0]

    load_res = [2, 12, 16, 17, 18, 19]
    load_ind = [22]
    load_com = [24, 35, 36, 37, 40, 41, 42, 43]
    sgen = load_res + load_ind + load_com
    tap_changer = {
        0: {"min": -10, "max": 10, "mid": 0, "ts_size": 0.625},
        1: {"min": -10, "max": 10, "mid": 0, "ts_size": 0.625},
    }

    for bus_id in load_res:
        pp.create_load(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="LOAD_RES_{}".format(bus_id),
            scaling=1.0,
            in_service=True,
            controllable=False,
        )

    for bus_id in load_ind:
        pp.create_load(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="LOAD_IND_{}".format(bus_id),
            scaling=1.0,
            in_service=True,
            controllable=False,
        )

    for bus_id in load_com:
        pp.create_load(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="LOAD_COM_{}".format(bus_id),
            scaling=1.0,
            in_service=True,
            controllable=False,
        )

    for bus_id in sgen:
        pp.create_sgen(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="SGEN_{}".format(bus_id),
            scaling=1.0,
            type=None,
            in_service=True,
            controllable=False,
        )

    for trafo_id, trafo_config in tap_changer.items():
        grid.trafo.loc[trafo_id, "tap_side"] = "lv"
        grid.trafo.loc[trafo_id, "tap_min"] = trafo_config["min"]
        grid.trafo.loc[trafo_id, "tap_max"] = trafo_config["max"]
        grid.trafo.loc[trafo_id, "tap_neutral"] = trafo_config["mid"]
        grid.trafo.loc[trafo_id, "tap_step_percent"] = trafo_config["ts_size"]

    return grid



import logging
from typing import List

import pandapower as pp

from ..constraints.voltage_band import ConstraintVoltageBand
from .base import GridElement

LOG = logging.getLogger(__name__)


class PPBus(GridElement):
    @staticmethod
    def pp_key() -> str:
        return "bus"

    @staticmethod
    def res_pp_key() -> str:
        return "res_bus"

    def __init__(self, index, grid, value):
        super().__init__(index, grid, LOG)

        self.in_service = True
        self.add_constraint(
            ConstraintVoltageBand(self, 1.0 - value, 1.0 + value)
        )
        self.load_indices: List[int] = []
        for idx, bus in zip(
            self.grid.get_value("load").index.values,
            self.grid.get_value("load", attr="bus").values,
        ):
            if bus == self.index:
                self.load_indices.append(int(idx))
        self.sgen_indices: List[int] = []
        for idx, bus in zip(
            self.grid.get_value("sgen").index.values,
            self.grid.get_value("sgen", attr="bus").values,
        ):
            if bus == self.index:
                self.sgen_indices.append(int(idx))

    def step(self, time):
        old_state = self.in_service
        self.in_service = True
        self.set_load_service_state(in_service=True)
        self.set_sgen_service_state(in_service=True)

        self._check(time)

        if old_state != self.in_service:
            if self.in_service:
                # Was off; but can switch on now?
                try:
                    self.grid.run_powerflow()
                    self._check(time)
                except pp.powerflow.LoadflowNotConverged:
                    self.set_load_service_state(in_service=False)
                    self.set_sgen_service_state(in_service=False)
                    self.in_service = False

        # Announce possible state changes
        if old_state != self.in_service:
            if not self.in_service:
                msg = (
                    f"At step {time}: Bus {self.index} out of service! "
                    f"(vm_pu: {self._constraints[0].violated_value}). "
                    f"Disabling loads {self.load_indices} and sgens "
                    f"{self.sgen_indices}"
                )
                LOG.debug(msg)
                try:
                    self.grid.run_powerflow()
                except pp.powerflow.LoadflowNotConverged:
                    LOG.debug("Bus disabled, LF still not converges.")
            else:
                LOG.debug(f"At step {time}: Bus {self.index} back in service!")
                try:
                    self.grid.run_powerflow()
                except pp.powerflow.LoadflowNotConverged:
                    LOG.debug("Bus re-enabled, LF not converging.")

        # We won't set the bus out of service; only the loads or sgens
        # connected to it
        # self.set_value("in_service", self.in_service)

        # Run powerflow so that the next unit does not automatically
        # switch off, too.

        return old_state != self.in_service

    def set_load_service_state(self, in_service: bool):
        for idx in self.load_indices:
            self.grid.set_value("load", idx, "in_service", in_service)

    def set_sgen_service_state(self, in_service: bool):
        for idx in self.sgen_indices:
            self.grid.set_value("sgen", idx, "in_service", in_service)



"""In this module, the midas mv grid is defined and configured."""

import pandapower as pp
import pandapower.networks as pn


def build_grid(**kwargs):
    """Create the default midas mv grid."""
    # DERs are added manually
    grid = pn.create_cigre_network_mv(with_der="all")
    grid.load = grid.load[0:0]
    grid.sgen = grid.sgen[0:0]
    grid.storage = grid.storage[0:0]

    load_res = [1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    load_com = [13, 14]
    sgen = [1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    storage = [1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    tap_changer = {
        0: {"min": -10, "max": 10, "mid": 0, "ts_size": 0.625},
        1: {"min": -10, "max": 10, "mid": 0, "ts_size": 0.625},
    }

    for bus_id in load_res:
        pp.create_load(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="LOAD_AGGRLV_{}".format(bus_id),
            scaling=1.0,
            in_service=True,
            controllable=False,
        )

    for bus_id in load_com:
        pp.create_load(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="LOAD_COMM_{}".format(bus_id),
            scaling=1.0,
            in_service=True,
            controllable=False,
        )

    for bus_id in sgen:
        pp.create_sgen(
            grid,
            bus=bus_id,
            p_mw=0.0,
            q_mvar=0.0,
            name="SGEN_{}".format(bus_id),
            scaling=1.0,
            type=None,
            in_service=True,
            controllable=False,
        )

    for bus_id in storage:
        pp.create_storage(
            grid,
            bus=bus_id,
            p_mw=0.0,
            max_e_mwh=1,
            q_mvar=0.0,
            name="STORAGE_{}".format(bus_id),
            scaling=1.0,
            type=None,
            in_service=True,
            controllable=False,
        )

    for trafo_id, trafo_config in tap_changer.items():
        grid.trafo.loc[trafo_id, "tap_side"] = "lv"
        grid.trafo.loc[trafo_id, "tap_min"] = trafo_config["min"]
        grid.trafo.loc[trafo_id, "tap_max"] = trafo_config["max"]
        grid.trafo.loc[trafo_id, "tap_neutral"] = trafo_config["mid"]
        grid.trafo.loc[trafo_id, "tap_step_percent"] = trafo_config["ts_size"]

    num_loads = len(load_res) + len(load_com)
    num_sgens = len(sgen)

    if "constant_load_p_mw" in kwargs:
        grid.load.p_mw.values[:] = [
            kwargs["constant_load_p_mw"] for _ in range(num_loads)
        ]
    if "constant_load_q_mvar" in kwargs:
        grid.load.q_mvar.values[:] = [
            kwargs["constant_load_q_mvar"] for _ in range(num_loads)
        ]

    if "constant_sgen_p_mw" in kwargs:
        grid.sgen.p_mw.values[:] = [
            kwargs["constant_sgen_p_mw"] for _ in range(num_sgens)
        ]

    if "constant_sgen_q_mvar" in kwargs:
        grid.sgen.q_mvar.values[:] = [
            kwargs["constant_sgen_q_mvar"] for _ in range(num_sgens)
        ]
    return grid



import unittest

from midas_powergrid.model.static import PandapowerGrid


class TestConstraintedGrid(unittest.TestCase):
    def test_line_constraints(self):
        gparams = {
            "constant_load_p_mw": 0.7,  # 0.5
            "constant_load_q_mvar": 0.23,  # 0.17
            "constant_sgen_p_mw": 0.4,  # 0.3
            "constant_sgen_q_mvar": -0.13,  # -0.1
            "constraints": [["line", 100]],
            "gridfile": "midasmv",
        }
        self.grid = PandapowerGrid(gparams)

        step_size = 60
        results = []
        setpoints = []
        for i in range(100):
            setpoints.append(0.7 + i * 0.05)
            self.grid.set_inputs("load", 1, {"p_mw": setpoints[-1]})
            # print(self.grid.grid.load.iloc[1]["p_mw"])
            self.grid.run_powerflow(i * step_size)
            results.append(self.grid.get_outputs())

        checks = []
        for i, res in enumerate(results):
            checks.append(
                self._assert_line_loading_below_limit(res, 100, setpoints[i])
            )
        # print(checks)

    def _assert_line_loading_below_limit(self, results, limit, p_set):
        lines = {k: v for k, v in results.items() if "line" in k}
        loadings = {
            line: val
            for line, data in lines.items()
            for key, val in data.items()
            if "loading" in key
        }
        checks = {}
        for line, loading in loadings.items():
            checks[line] = loading <= limit

            self.assertLessEqual(
                loading,
                limit,
                msg=f"Loading of line {line} is above limit: {loading} "
                f"/ {limit}",
            )

        # print(loadings)
        return checks

    @unittest.skip
    def test_line_disables_after_violation(self):
        self.grid = PandapowerGrid()
        gparams = {
            "constant_load_p_mw": 0.7,  # 0.5
            "constant_load_q_mvar": 0.23,  # 0.17
            "constant_sgen_p_mw": 0.4,  # 0.3
            "constant_sgen_q_mvar": -0.13,  # -0.1
            "constraints": [["line", 100]],
        }
        self.grid.setup("midasmv", 0, gparams)
        self.grid.set_inputs("load", 1, {"p_mw": 1.5})
        self.grid.run_powerflow(0)
        # results = self.grid.get_outputs()
        # print()

    def test_voltage_constraints(self):
        gparams = {
            "constant_load_p_mw": 0.7,  # 0.5
            "constant_load_q_mvar": 0.23,  # 0.17
            "constant_sgen_p_mw": 0.4,  # 0.3
            "constant_sgen_q_mvar": -0.13,  # -0.1
            "gridfile": "midasmv",
        }
        self.ugrid = PandapowerGrid(gparams)

        gparams["constraints"] = [["sgen", 0.05], ["load", 0.02], ["bus", 0.1]]
        self.grid = PandapowerGrid(gparams)
        step_size = 60

        results = []
        uresults = []
        setpoints = []
        for i in range(100):
            setpoints.append(0.7 + i * 0.05)
            self.grid.set_inputs("load", 1, {"p_mw": setpoints[-1]})
            self.grid.run_powerflow(i * step_size)
            results.append(self.grid.get_outputs())

            self.ugrid.set_inputs("load", 1, {"p_mw": setpoints[-1]})
            self.ugrid.run_powerflow(i * step_size)
            uresults.append(self.ugrid.get_outputs())

        checks = []
        for i, (ures, res) in enumerate(zip(uresults, results)):
            checks.append(
                self._assert_voltage_within_limits(
                    ures, res, 0.9, 1.1, setpoints[i]
                )
            )
        # print(checks)

    def _assert_voltage_within_limits(
        self, uresults, results, low, high, p_set
    ):
        ubuses = {k: v for k, v in uresults.items() if "bus" in k}
        uvm_pus = {
            bus: val
            for bus, data in ubuses.items()
            for key, val in data.items()
            if "vm_pu" in key
        }
        buses = {k: v for k, v in results.items() if "bus" in k}
        vm_pus = {
            bus: val
            for bus, data in buses.items()
            for key, val in data.items()
            if "vm_pu" in key
        }

        checks = {}
        for bus in uvm_pus:
            vm_pu = uvm_pus[bus]
            checks[bus] = low <= vm_pu <= high

            if not checks[bus]:
                self.assertLessEqual(
                    vm_pus[bus],
                    high,
                    msg=(
                        f"Voltage of bus {bus} is above limit: {vm_pus[bus]} "
                        f"/ {high}",
                    ),
                )
                self.assertGreaterEqual(
                    vm_pus[bus],
                    low,
                    msg=(
                        f"Voltage of bus {bus} is below limit: {vm_pus[bus]} "
                        f"/ {low}",
                    ),
                )
        return checks


if __name__ == "__main__":
    unittest.main()



from .base import Constraint


class ConstraintTapChanges(Constraint):
    def __init__(self, element, expected_value):
        super().__init__(element)

        self._changes_per_hour = expected_value
        self._last_change = 0
        self._changes = dict()

    def check(self, time) -> bool:
        self._satisfied = True
        current_pos = self._element.grid[self._element.pp_key()][
            "tap_pos"
        ].loc[self._element.index]

        if current_pos != self._element.tap_pos:
            self._changes[time] = self._element.tap_pos

            start = max(0, time - 3600)
            num_changes = 0
            for t in range(start, time):
                if t in self._changes:
                    num_changes += 1

            if num_changes > self._changes_per_hour:
                self._satisfied = False

        return self._satisfied

    def handle_violation(self):
        self._element.tap_pos = self._element.grid[self._element.pp_key][
            "tap_pos"
        ].loc[self._element.index]



from .voltage_change import ConstraintVoltageChange


class ConstraintLoadVoltageChange(ConstraintVoltageChange):
    """Voltage change of a load is not allowed to rise over 2%."""

    def __init__(self, element, expected_value):
        super().__init__(element)

        self._expected_value = expected_value

    def handle_violation(self):
        # Overvoltages should be handles by generators
        if self._time_voltages[-1].value < 1.0:
            self._element.in_service = False

        self._element.current_bus_voltage = self._time_voltages[-1].value



"""This module contains the mosaik meta definition for the
pandapower simulator.

"""

import numpy as np

ATTRIBUTE_MAP = {
    "Bus": {
        "bus": [
            ("in_service", "bool", 0, 2),  # bus is operational
            ("vn_kv", "float", 0, 440),  # nominal voltage
        ],
        "res_bus": [
            ("p_mw", "float", -np.inf, np.inf),  # active power
            ("q_mvar", "float", -np.inf, np.inf),  # reactive power
            ("vm_pu", "float", 0.0, np.inf),  # voltage magnitude
            ("va_degree", "float", -180.0, 180.0),  # voltage angle
        ],
    },
    "Line": {
        "line": [
            ("in_service", "bool", 0, 2),  # line is operational
            ("max_i_ka", "float", 0, np.inf),  # maximum current
        ],
        "res_line": [
            ("loading_percent", "float", 0, np.inf),  # utilization
            ("p_from_mw", "float", -np.inf, np.inf),  # active power from
            ("p_to_mw", "float", -np.inf, np.inf),  # active power to
            ("q_from_mvar", "float", -np.inf, np.inf),  # reactive power from
            ("q_to_mvar", "float", -np.inf, np.inf),  # reactive power to
            ("vm_from_pu", "float", 0.0, np.inf),  # voltage magnitude from
            ("vm_to_pu", "float", 0.0, np.inf),  # voltage magnitude to
            ("va_from_degree", "float", -180.0, 180.0),  # voltage angle from
            ("va_to_degree", "float", -180.0, 180.0),  # voltage angle to
            ("i_from_ka", "float", -np.inf, np.inf),  # current from
            ("i_to_ka", "float", -np.inf, np.inf),  # current to
        ],
    },
    "Trafo": {
        "trafo": [
            ("in_service", "bool", 0, 2),  # trafo is operational
            ("tap_pos", "float", -1.0, 1.0),  # position of tap changer
            ("tap_min", "int", -np.inf, np.inf),  # minimal tap position
            ("tap_max", "int", -np.inf, np.inf),  # maximum tap position
        ],
        "res_trafo": [
            ("loading_percent", "float", 0, np.inf),  # utilization
            ("p_hv_mw", "float", -np.inf, np.inf),  # active power hv
            ("p_lv_mw", "float", -np.inf, np.inf),  # active power lv
            ("q_hv_mvar", "float", -np.inf, np.inf),  # reactive power hv
            ("q_lv_mvar", "float", -np.inf, np.inf),  # reactive power lv
            ("vm_hv_pu", "float", 0.0, np.inf),  # voltage magnitude hv
            ("vm_lv_pu", "float", 0.0, np.inf),  # voltage magnitude lv
            ("va_hv_degree", "float", -180.0, 180.0),  # voltage angle hv
            ("va_lv_degree", "float", -180.0, 180.0),  # voltage angle lv
            ("i_hv_ka", "float", -np.inf, np.inf),  # current hv
            ("i_lv_ka", "float", -np.inf, np.inf),  # current lv
        ],
    },
    "Switch": {
        "switch": [("closed", "float", 0, 1.0)],
        "res_switch": [("i_ka", "float", -np.inf, np.inf)],
    },
    "Ext_grid": {
        "ext_grid": [],
        "res_ext_grid": [
            ("p_mw", "float", -np.inf, np.inf),  # active power
            ("q_mvar", "float", -np.inf, np.inf),  # reactive power
        ],
    },
    "Load": {
        "load": [
            ("p_mw", "float", -np.inf, np.inf),  # active power
            ("q_mvar", "float", -np.inf, np.inf),  # reactive power
            ("in_service", "bool", 0, 2),  # load is operational
        ],
        "res_load": [],
    },
    "Sgen": {
        "sgen": [
            ("p_mw", "float", -np.inf, np.inf),  # active power
            ("q_mvar", "float", -np.inf, np.inf),  # reactive power
            ("in_service", "bool", 0, 2),  # generator is operational
        ],
        "res_sgen": [],
    },
    "Storage": {
        "storage": [
            ("p_mw", "float", -np.inf, np.inf),  # active power
            ("q_mvar", "float", -np.inf, np.inf),  # reactive power
            ("in_service", "bool", 0, 2),  # storage is operational
        ],
        "res_storage": [],
    },
}
META = {
    "type": "time-based",
    "models": {
        "Grid": {
            "public": True,
            "params": [
                "gridfile",  # Name of the grid topology
                "pp_params",
                "plotting",  # Flag to activate plotting
                "use_constraints",  # Flag to activate constraints
                "constraints",  # A list of constraints to activate
                "actuator_multiplier",
                "include_slack_bus",
            ],
            "attrs": ["health", "grid_json"],
        },
        "Ext_grid": {
            "public": False,
            "params": [],
            "attrs": [
                attr[0] for attr in ATTRIBUTE_MAP["Ext_grid"]["ext_grid"]
            ]
            + [attr[0] for attr in ATTRIBUTE_MAP["Ext_grid"]["res_ext_grid"]],
            # "attrs": [
            #     "p_mw",  # load active power [MW]
            #     "q_mvar",  # load reactive power [MVAr]
            # ],
        },
        "Bus": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Bus"]["bus"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Bus"]["res_bus"]],
            # "attrs": [
            #     "p_mw",  # load Active power [MW]
            #     "q_mvar",  # Reactive power [MVAr]
            #     "vn_kv",  # Nominal bus voltage [KV]
            #     "vm_pu",  # Voltage magnitude [p.u]
            #     "va_degree",  # Voltage angle [deg]
            #     "in_service",  # Bus is in service [bool]
            # ],
        },
        "Load": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Load"]["load"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Load"]["res_load"]],
            # "attrs": [
            #     "p_mw",  # load Active power [MW]
            #     "q_mvar",  # Reactive power [MVAr]
            #     "in_service",  # specifies if the load is in service.
            #     "controllable",  # States if load is controllable or not.
            # ],
        },
        "Sgen": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Sgen"]["sgen"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Sgen"]["res_sgen"]],
            # "attrs": [
            #     "p_mw",  # load Active power [MW]
            #     "q_mvar",  # Reactive power [MVAr]
            #     "in_service",  # specifies if the load is in service.
            #     "controllable",  # States if load is controllable or not.
            #     "va_degree",  # Voltage angle [deg]
            # ],
        },
        "Trafo": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Trafo"]["trafo"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Trafo"]["res_trafo"]],
            # "attrs": [
            #     "p_hv_mw",  # Active power at "from" side [MW]
            #     "q_hv_mvar",  # Reactive power at "from" side [MVAr]
            #     "p_lv_mw",  # Active power at "to" side [MW]
            #     "q_lv_mvar",  # Reactive power at "to" side [MVAr]
            #     "sn_mva",  # Rated apparent power [MVA]
            #     "max_loading_percent",  # Maximum Loading
            #     "vn_hv_kv",  # Nominal primary voltage [kV]
            #     "vn_lv_kv",  # Nominal secondary voltage [kV]
            #     "vm_hv_pu",
            #     "vm_lv_pu",
            #     "va_hv_degree",
            #     "va_lv_degree",
            #     "pl_mw",  # Active power loss [MW]
            #     "ql_mvar",  # reactive power consumption of the
            #     # transformer [Mvar]
            #     # 'pfe_kw',       #  iron losses in kW [kW]
            #     # 'i0_percent',       #  iron losses in kW [kW]
            #     "loading_percent",  # load utilization relative to rated
            #     # power [%]
            #     "i_hv_ka",  # current at the high voltage side of the
            #     # transformer [kA]
            #     "i_lv_ka",  # current at the low voltage side of the
            #     # transformer [kA]
            #     "tap_max",  # maximum possible  tap turns
            #     "tap_min",  # minimum possible tap turns
            #     "tap_pos",  # Currently active tap turn
            #     "in_service",  # Specifies if the trafo is in service
            # ],
        },
        "Line": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Line"]["line"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Line"]["res_line"]],
            # "attrs": [
            #     "p_from_mw",  # Active power at "from" side [MW]
            #     "q_from_mvar",  # Reactive power at "from" side [MVAr]
            #     "p_to_mw",  # Active power at "to" side [MW]
            #     "q_to_mvar",  # Reactive power at "to" side [MVAr]
            #     "max_i_ka",  # Maximum current [KA]
            #     "length_km",  # Line length [km]
            #     "pl_mw",  # active power losses of the line [MW]
            #     "ql_mvar",  # reactive power consumption of the line [MVar]
            #     "i_from_ka",  # Current at from bus [kA]
            #     "i_to_ka",  # Current at to bus [kA]
            #     "loading_percent",  # line loading [%]
            #     "r_ohm_per_km",  # Resistance per unit length [Ω/km]
            #     "x_ohm_per_km",  # Reactance per unit length [Ω/km]
            #     "c_nf_per_km",  # Capactity per unit length [nF/km]
            #     "in_service",  # Boolean flag (True|False)
            #     "vm_from_pu",
            #     "vm_to_pu",
            #     "va_from_degree",
            #     "va_to_degree",
            # ],
        },
        "Switch": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Switch"]["switch"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Switch"]["res_switch"]],
            # "attrs": [
            #     "et",
            #     "type",
            #     "closed",
            # ],
        },
        "Storage": {
            "public": False,
            "params": [],
            "attrs": [attr[0] for attr in ATTRIBUTE_MAP["Storage"]["storage"]]
            + [attr[0] for attr in ATTRIBUTE_MAP["Storage"]["res_storage"]],
            # "attrs": [
            #     "p_mw",  # load Active power [MW]
            #     "q_mvar",  # Reactive power [MVAr]
            #     "max_e_mwh",  # maximum charge level
            #     "in_service",  # specifies if the load is in service.
            #     "controllable",  # States if load is controllable or not.
            # ],
        },
    },
}



import os

import matplotlib.pyplot as plt
import numpy as np
from natsort import natsorted


def analyze_buses(data, report_file, name, output_path, full_report):
    data = data[[col for col in data.columns if "vm_pu" in col]]
    worst = {
        "voltage_dev": 0,
        "voltage_vio": 0,
        "buses": {},
    }
    best = {
        "voltage_dev": float("inf"),
        "voltage_vio": float("inf"),
        "buses": {},
    }
    fails = 0
    total = 0

    report_file.append("## Bus Analysis\n")
    data = data[natsorted(data.columns)]
    for colo in data.columns:
        col = colo.replace("___", ".").replace("__", "-")
        vm_pus = data[colo].values
        key = col.split(".")[0]
        len_bus = len(vm_pus)

        total += len_bus
        voltage_vio = _voltage_violation_report(vm_pus, key, report_file)
        fails += voltage_vio
        voltage_dev = _voltage_deviation(vm_pus)

        if voltage_vio >= worst["voltage_vio"]:
            if voltage_dev > worst["voltage_dev"]:
                worst["buses"].clear()
                update_extremum(worst, vm_pus, key, voltage_vio, voltage_dev)
            elif voltage_dev == worst["voltage_dev"]:
                update_extremum(worst, vm_pus, key, voltage_vio, voltage_dev)

        if voltage_vio <= best["voltage_vio"]:
            if voltage_dev < best["voltage_dev"]:
                best["buses"].clear()
                update_extremum(best, vm_pus, key, voltage_vio, voltage_dev)
            elif voltage_dev == best["voltage_dev"]:
                update_extremum(best, vm_pus, key, voltage_vio, voltage_dev)

        if full_report:
            full_analyze(name, output_path, vm_pus, key)

    if not full_report:
        for key, vm_pus in worst["buses"].items():
            full_analyze(
                name, output_path, vm_pus, "worst_" + key, report_file
            )
        for key, vm_pus in best["buses"].items():
            full_analyze(name, output_path, vm_pus, "best_" + key, report_file)

    report_file.append("")

    aggregated = data.mean(axis=1).values
    annual = np.sort(aggregated)[::-1]
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 8))
    ax1.plot(aggregated)
    ax1.axhline(y=1.1, color="red")
    ax1.axhline(y=1.04, linestyle="--", color="red")
    ax1.axhline(y=0.96, linestyle="--", color="red")
    ax1.axhline(y=0.9, color="red")
    ax1.set_title("0-buses vm_pu")

    ax1.set_ylabel("voltage magnitude p.u.")
    ax2.plot(annual)
    ax2.axhline(y=1.1, color="red")
    ax2.axhline(y=1.04, linestyle="--", color="red")
    ax2.axhline(y=0.96, linestyle="--", color="red")
    ax2.axhline(y=0.9, color="red")
    ax2.set_title("0-buses annual curve vm_pu")
    ax2.set_xlabel("time (15-minute steps)")
    ax2.set_ylabel("voltage magnitude p.u.")
    img_path = os.path.join(output_path, f"{name}_0-buses_vmpu.png")
    plt.savefig(img_path, dpi=300, bbox_inches="tight")
    plt.close()

    report_file.append(f"\n![Total_Bus_vm_pu]({img_path})" + "{width=100%}\n")

    score = 100 * (1.0 - (fails / total))
    return score


def update_extremum(categorie, vm_pus, key, voltage_vio, voltage_dev):
    categorie["voltage_dev"] = voltage_dev
    categorie["voltage_vio"] = voltage_vio
    categorie["buses"][key] = vm_pus


def full_analyze(name, output_path, vm_pus, key, report_file=None):
    annual = np.sort(vm_pus)[::-1]
    fig, axes = plt.subplots(2, 1, figsize=(9, 9))
    for ax, series, title in zip(
        axes, [vm_pus, annual], [f"{key} vm_pu", f"{key} vm_pu annual"]
    ):
        ax.plot(series)
        ax.axhline(y=1.1, color="red")
        ax.axhline(y=1.04, linestyle="--", color="red")
        ax.axhline(y=0.96, linestyle="--", color="red")
        ax.axhline(y=0.9, color="red")

        ax.set_title(title)
        ax.set_ylabel("voltage magnitude p.u.")
        ax.set_xlabel("time (15 minute steps)")

    img_path = os.path.join(output_path, f"{name}_{key}_vmpu.png")
    plt.savefig(img_path, dpi=300, bbox_inches="tight")
    plt.close()
    if report_file is not None:
        categorie = key.split()[0]
        report_file.append(
            f"\n![{categorie}_Bus_vm_pu]({img_path})" + "{width=100%}\n"
        )


def _voltage_violation_report(vm_pus, key, report_file):
    too_high10 = (vm_pus > 1.1).sum()
    too_high4 = (vm_pus > 1.04).sum()
    too_low10 = (vm_pus < 0.9).sum()
    too_low4 = (vm_pus < 0.96).sum()

    if too_high10 > 0:
        score = 100 * too_high10 / len(vm_pus)
        report_file.append(
            f"* [{key}] {too_high10} values > 1.1 ({score:.3f} %)"
        )
    if too_high4 - too_high10 > 0:
        score = 100 * (too_high4 - too_high10) / len(vm_pus)
        report_file.append(
            f"* [{key}] {too_high4-too_high10} values > 1.04 ({score:.3f} %)"
        )
    if too_low10 > 0:
        score = 100 * too_low10 / len(vm_pus)
        report_file.append(
            f"* [{key}] {too_low10} values < 0.9 ({score:.3f} %)"
        )
    if too_low4 - too_low10 > 0:
        score = 100 * (too_low4 - too_low10) / len(vm_pus)
        report_file.append(
            f"* [{key}] {too_low4-too_low10} values < 0.96 ({score:.3f} %)"
        )

    return too_low4 + too_high4


def _voltage_deviation(vm_pus):
    return np.abs(vm_pus - 1).sum()



"""Creates a pandapower network of the BHV power grid."""

import pandapower as pp
from midas.util.compute_q import compute_q


def build_grid(
    invert_geodata: bool = False, use_extension: bool = False, **kwargs
) -> pp.pandapowerNet:
    """Return a pandapower network of BHV based on FLOSM map data.

    Returns
    -------
    pp.pandapowerNet
        powergrid model of Bremerhaven.

    """
    grid = pp.create_empty_network()
    buses = dict()
    _create_line_types(grid)
    _create_buses_hv(grid, buses)
    _create_lines_hv(grid, buses)
    _create_overseas(grid, buses)
    _create_lehe(grid, buses)
    _create_windpark_speckenbuettel(grid, buses)
    _create_windpark_multibrid(grid, buses)
    _create_geestemuende(grid, buses)
    _create_city(grid, buses, use_extension)
    _create_windpark_lehe(grid, buses)
    _create_klinikum_reinkenheide(grid, buses)
    _create_schiffdorferdamm(grid, buses)
    _create_eisarena(grid, buses)
    _create_mitte(grid, buses)
    _create_leherheide(grid, buses)
    _create_wulsdorf(grid, buses)
    _create_abfall(grid, buses, use_extension)
    _create_surheide(grid, buses)
    _create_klinikum_buergerpark(grid, buses)
    _create_innenstadt(grid, buses)
    _create_zoo(grid, buses)
    _create_klinikum_mitte(grid, buses)
    _create_fischereihafen(grid, buses)
    _create_geestemuende_v2(grid, buses)
    _create_bremerhaven_sued(grid, buses)

    if invert_geodata:
        _invert_geodata(grid)

    return grid


def _create_line_types(grid):
    """Add line data based on CIGRE."""
    line_data = {
        "c_nf_per_km": 151.1749,
        "r_ohm_per_km": 0.501,
        "x_ohm_per_km": 0.716,
        "max_i_ka": 0.145,
        "type": "cs",
    }
    pp.create_std_type(grid, line_data, name="CABLE_CIGRE_MV", element="line")

    line_data = {
        "c_nf_per_km": 10.09679,
        "r_ohm_per_km": 0.510,
        "x_ohm_per_km": 0.366,
        "max_i_ka": 0.195,
        "type": "ol",
    }
    pp.create_std_type(grid, line_data, name="OHL_CIGRE_MV", element="line")

    line_data = {
        "c_nf_per_km": 0.0,
        "r_ohm_per_km": 0.162,
        "x_ohm_per_km": 0.0832,
        "max_i_ka": 1.0,
        "type": "cs",
    }
    pp.create_std_type(grid, line_data, name="UG1", element="line")


def _create_buses_hv(net, buses):
    """Instantiate 110 kV buses based on FLOSM geodata"""
    # Nordenham Blexen
    # https://www.openstreetmap.org/way/37932525
    buses["hv_bus_0"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Umspannwerk Blexen/hv_bus_0",
        type="b",
        geodata=(53.528571, 8.545029),
    )

    buses["ov_hv_0"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Überseehafen, Wurster Straße/ov_hv_0",
        type="b",
        geodata=(53.597675, 8.538051),
    )

    # https://www.openstreetmap.org/way/107058988
    buses["hv_bus_2"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Überseehafen, Grauwallring/hv_bus_2",
        type="b",
        geodata=(53.596534, 8.571337),
    )
    # Nord - Leherheide
    # https://www.openstreetmap.org/way/41831269
    buses["hv_bus_3"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Leherheide, Debstedter Weg/hv_bus_3",
        type="b",
        geodata=(53.602492, 8.621490),
    )
    # Mitte - Lehe
    # https://www.openstreetmap.org/way/111990696
    buses["hv_bus_4"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Lehe, Am Wischhacker/hv_bus_4",
        type="b",
        geodata=(53.560075, 8.601626),
    )
    # Mitte - Schiffdorferdamm
    # https://www.openstreetmap.org/node/1195969029
    buses["hv_bus_5"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Schiffdorferdamm, Johann-Wichels-Weg/hv_bus_5",
        type="b",
        geodata=(53.539539, 8.629766),
    )
    # Süd - Wulsdorf
    # https://www.openstreetmap.org/way/117216153
    buses["hv_bus_6"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Surheide, Midgardweg/hv_bus_6",
        type="b",
        geodata=(53.510913, 8.620990),
    )
    # https://www.openstreetmap.org/way/36903323
    buses["hv_bus_7"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Wulsdorf, Daimlerstraße/hv_bus_7",
        type="b",
        geodata=(53.515625, 8.603047),
    )
    # https://www.openstreetmap.org/way/370348954
    buses["hv_bus_8"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Wulsdorf, Hackfahrel/hv_bus_8",
        type="b",
        geodata=(53.500857, 8.595032),
    )
    # Mitte - Kaiserhafen Eins
    # https://www.openstreetmap.org/node/6470789771
    buses["hv_bus_9"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Kaiserhafen Eins, Lohmannstraße/hv_bus_9",
        type="b",
        geodata=(53.555433, 8.560829),
    )
    # Mitte - Mitte
    # https://www.openstreetmap.org/way/252421504
    buses["hv_bus_10"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Mitte, Barkhausenstraße/hv_bus_10",
        type="b",
        geodata=(53.548044, 8.572743),
    )
    # Mitte - Geestemünde-Nord
    # https://www.openstreetmap.org/node/2335975543
    buses["hv_bus_11"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Geestemünde-Nord, Mozartstraße/hv_bus_11",
        type="b",
        geodata=(53.543114, 8.593258),
    )
    # https://www.openstreetmap.org/way/292524887
    buses["hv_bus_12"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Handelshafen, Klußmannstraße/hv_bus_12",
        type="b",
        geodata=(53.532868, 8.583921),
    )


def _create_lines_hv(net, buses):
    """
    Connect buses based on FLOSM power pole geodata
    """
    pp.create_line(
        net,
        from_bus=buses["ov_hv_0"],
        to_bus=buses["hv_bus_3"],
        length_km=9.96,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.597675, 8.538051),  # from_bus
            (53.598356, 8.538919),
            (53.598247, 8.540951),
            (53.598018, 8.544490),
            (53.598765, 8.548444),
            (53.599555, 8.552571),
            (53.600899, 8.555071),
            (53.602217, 8.557474),
            (53.601375, 8.561635),
            (53.600501, 8.565854),
            (53.599581, 8.570314),
            (53.598819, 8.573943),
            (53.600817, 8.577339),
            (53.602896, 8.580954),
            (53.604454, 8.583614),
            (53.606914, 8.584638),
            (53.609244, 8.585576),
            (53.611986, 8.586714),
            (53.614576, 8.587774),
            (53.617890, 8.589153),
            (53.620520, 8.592915),
            (53.623156, 8.596658),
            (53.623434, 8.601912),
            (53.623702, 8.606821),
            (53.622806, 8.611870),
            (53.622035, 8.616256),
            (53.621311, 8.620426),
            (53.620383, 8.625599),
            (53.617257, 8.628189),
            (53.614890, 8.628631),
            (53.611929, 8.629195),
            (53.609874, 8.627677),
            (53.607835, 8.626151),
            (53.605222, 8.624216),
            (53.602965, 8.622550),
            (53.602492, 8.621490),  # to_bus
        ],
    )
    buses["c_hv_0"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Langen Bremerhaven Nord/c_hv_0",
        type="b",
        geodata=(53.60537, 8.6241),
    )
    # hv bzs 5
    pp.create_line(
        net,
        from_bus=buses["hv_bus_5"],
        to_bus=buses["c_hv_0"],
        length_km=2,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            _bus_geo(net, buses["hv_bus_5"]),
            _bus_geo(net, buses["c_hv_0"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_3"],
        to_bus=buses["hv_bus_4"],
        length_km=10.08,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.602492, 8.621490),  # from_bus
            (53.602718, 8.623000),
            (53.604402, 8.625998),
            (53.604395, 8.630796),
            (53.604373, 8.636187),
            (53.604380, 8.641501),
            (53.604358, 8.647265),
            (53.604344, 8.651771),
            (53.604326, 8.657091),
            (53.601005, 8.656239),
            (53.597519, 8.655461),
            (53.593976, 8.654641),
            (53.590606, 8.653878),
            (53.586869, 8.653052),
            (53.583463, 8.652269),
            (53.580202, 8.651513),
            (53.576792, 8.650749),
            (53.573688, 8.650031),
            (53.570436, 8.649304),
            (53.567128, 8.648553),
            (53.564986, 8.645384),
            (53.562641, 8.641858),
            (53.560282, 8.638338),
            (53.558868, 8.634269),
            (53.558998, 8.632271),
            (53.559300, 8.627266),
            (53.559599, 8.622411),
            (53.559884, 8.617658),
            (53.560164, 8.612923),
            (53.560471, 8.607664),
            (53.560258, 8.602617),
            (53.560075, 8.601626),  # to_bus
        ],
    )
    buses["c_hv_1"] = pp.create_bus(
        net,
        vn_kv=110,
        name="110 kV Lehe/c_hv_1",
        type="b",
        geodata=(53.563454, 8.601023),
    )

    pp.create_line(
        net,
        from_bus=buses["hv_bus_4"],
        to_bus=buses["c_hv_1"],
        length_km=2,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            _bus_geo(net, buses["hv_bus_4"]),
            _bus_geo(net, buses["c_hv_1"]),
        ],
    )

    pp.create_line(
        net,
        from_bus=buses["hv_bus_4"],
        to_bus=buses["hv_bus_6"],
        length_km=7.67,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.560075, 8.601626),  # from_bus
            (53.559832, 8.603161),
            (53.560022, 8.607649),
            (53.559732, 8.612116),
            (53.559403, 8.617469),
            (53.559058, 8.622976),
            (53.558708, 8.628350),
            (53.558359, 8.633821),
            (53.555374, 8.635590),
            (53.552654, 8.634057),
            (53.549372, 8.632231),
            (53.546590, 8.630663),
            (53.543353, 8.628849),
            (53.540546, 8.627274),
            (53.537848, 8.625771),
            (53.535248, 8.624267),
            (53.532523, 8.622777),
            (53.530186, 8.621398),
            (53.527832, 8.620009),
            (53.524874, 8.618278),
            (53.521592, 8.620127),
            (53.518242, 8.619302),
            (53.515789, 8.618684),
            (53.513301, 8.618067),
            (53.512171, 8.619612),
            (53.510913, 8.620990),  # to_bus
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_7"],
        to_bus=buses["hv_bus_6"],
        length_km=1.27,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.515625, 8.603047),  # from_bus
            (53.515551, 8.603400),
            (53.515384, 8.605437),
            (53.515142, 8.607763),
            (53.514873, 8.610541),
            (53.513592, 8.613797),
            (53.512418, 8.616813),
            (53.511601, 8.618849),
            (53.510913, 8.620990),  # to_bus
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_6"],
        to_bus=buses["hv_bus_8"],
        length_km=4.41,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.510913, 8.620990),  # from_bus
            (53.509886, 8.621935),
            (53.507894, 8.626273),
            (53.505894, 8.627324),
            (53.503911, 8.624381),
            (53.501855, 8.621282),
            (53.499356, 8.617602),
            (53.497031, 8.614142),
            (53.494530, 8.610403),
            (53.492585, 8.607504),
            (53.490529, 8.603516),
            (53.491313, 8.599338),
            (53.493578, 8.597337),
            (53.496759, 8.596275),
            (53.499637, 8.595319),
            (53.500857, 8.595032),  # to_bus
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_3"],
        to_bus=buses["hv_bus_6"],
        length_km=13.55,
        std_type="243-AL1/39-ST1A 110.0",
        geodata=[
            (53.602492, 8.621490),  # from_bus
            (53.602762, 8.622957),
            (53.602762, 8.622957),
            (53.604383, 8.630787),
            (53.604381, 8.636158),
            (53.604368, 8.641503),
            (53.604368, 8.647260),
            (53.604337, 8.651789),
            (53.604326, 8.656977),
            (53.526075, 8.639376),
            (53.524229, 8.635588),
            (53.523043, 8.633121),
            (53.521484, 8.629963),
            (53.519890, 8.626781),
            (53.518433, 8.623689),
            (53.516985, 8.620731),
            (53.515793, 8.618661),
            (53.513261, 8.617566),
            (53.511989, 8.619317),
            (53.510913, 8.620990),  # to_bus
        ],
    )

    # Connect remaining HV buses with hypothetical underground lines
    pp.create_line(
        net,
        from_bus=buses["ov_hv_0"],
        to_bus=buses["hv_bus_2"],
        length_km=2.19,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["ov_hv_0"]),
            _bus_geo(net, buses["hv_bus_2"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_4"],
        to_bus=buses["hv_bus_5"],
        length_km=2.9,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_4"]),
            _bus_geo(net, buses["hv_bus_5"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["ov_hv_0"],
        to_bus=buses["hv_bus_9"],
        length_km=4.93,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["ov_hv_0"]),
            _bus_geo(net, buses["hv_bus_9"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_9"],
        to_bus=buses["hv_bus_10"],
        length_km=1.14,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_9"]),
            _bus_geo(net, buses["hv_bus_10"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_10"],
        to_bus=buses["hv_bus_11"],
        length_km=1.47,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_10"]),
            _bus_geo(net, buses["hv_bus_11"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_11"],
        to_bus=buses["hv_bus_12"],
        length_km=1.3,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_11"]),
            _bus_geo(net, buses["hv_bus_12"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_12"],
        to_bus=buses["hv_bus_8"],
        length_km=3.63,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_12"]),
            _bus_geo(net, buses["hv_bus_8"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_0"],
        to_bus=buses["hv_bus_12"],
        length_km=2.61,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_0"]),
            _bus_geo(net, buses["hv_bus_12"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_7"],
        to_bus=buses["hv_bus_8"],
        length_km=1.73,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_7"]),
            _bus_geo(net, buses["hv_bus_8"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_7"],
        to_bus=buses["hv_bus_12"],
        length_km=2.29,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_7"]),
            _bus_geo(net, buses["hv_bus_12"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_10"],
        to_bus=buses["hv_bus_12"],
        length_km=1.84,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_10"]),
            _bus_geo(net, buses["hv_bus_12"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_2"],
        to_bus=buses["hv_bus_3"],
        length_km=3.39,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_2"]),
            _bus_geo(net, buses["hv_bus_3"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_2"],
        to_bus=buses["hv_bus_4"],
        length_km=4.52,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_2"]),
            _bus_geo(net, buses["hv_bus_4"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_4"],
        to_bus=buses["hv_bus_11"],
        length_km=1.97,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_4"]),
            _bus_geo(net, buses["hv_bus_11"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_5"],
        to_bus=buses["hv_bus_11"],
        length_km=2.44,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_5"]),
            _bus_geo(net, buses["hv_bus_11"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_4"],
        to_bus=buses["hv_bus_9"],
        length_km=2.74,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_4"]),
            _bus_geo(net, buses["hv_bus_9"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_3"],
        to_bus=buses["hv_bus_4"],
        length_km=4.00,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_3"]),
            _bus_geo(net, buses["hv_bus_4"]),
        ],
    )
    pp.create_line(
        net,
        from_bus=buses["hv_bus_5"],
        to_bus=buses["hv_bus_7"],
        length_km=3.19,
        std_type="N2XS(FL)2Y 1x240 RM/35 64/110 kV",
        geodata=[
            _bus_geo(net, buses["hv_bus_5"]),
            _bus_geo(net, buses["hv_bus_7"]),
        ],
    )


def _create_overseas(net, buses):
    pp.create_ext_grid(net, bus=buses["ov_hv_0"], name="Überseehafen")
    # Nord - Überseehafen
    # https://www.openstreetmap.org/way/157422539

    buses["ov_load_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Weddewarden Industrielast/ov_load_0",
        type="n",
        geodata=(53.604462, 8.534789),
    )

    pp.create_load(
        net,
        bus=buses["ov_load_0"],
        p_mw=0.06,
        name="Weddewarden Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.06),
    )
    pp.create_transformer(
        net,
        buses["ov_hv_0"],
        buses["ov_load_0"],
        name="110kV/20kV transformer",
        std_type="25 MVA 110/20 kV",
    )

    buses["ov_sgen_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Windpark Speckenbüttel 2/ov_sgen_0",
        type="n",
        geodata=(53.593788, 8.555553),
    )

    pp.create_sgen(
        net,
        bus=buses["ov_sgen_0"],
        p_mw=5.0,
        name="Windpark Speckenbüttel 2",
        type="WP",
        min_p_mw=0,
        max_p_mw=5.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=5),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["ov_load_0"],
        to_bus=buses["ov_sgen_0"],
        length_km=1.815,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["ov_load_0"]),
            _bus_geo(net, buses["ov_sgen_0"]),
        ],
    )

    buses["ov_lv_0"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Weddewarden LV/ov_lv_0",
        type="n",
        geodata=(53.604462, 8.534789),
    )

    pp.create_load(
        net,
        bus=buses["ov_lv_0"],
        p_mw=0.237,
        name="Weddewarden Households",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.237),
    )
    pp.create_sgen(
        net,
        bus=buses["ov_lv_0"],
        p_mw=0.03,
        name="PV Weddewarden",
        min_p_mw=0,
        max_p_mw=0.29,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.03),
    )
    buses["ov_load_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Weddewarden MV/ov_load_1",
        type="n",
        geodata=(53.604469, 8.534780),
    )

    pp.create_transformer_from_parameters(
        net,
        buses["ov_load_1"],
        buses["ov_lv_0"],
        sn_mva=0.4,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo-OS",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["ov_sgen_0"],
        to_bus=buses["ov_load_1"],
        length_km=0.01,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["ov_sgen_0"]),
            _bus_geo(net, buses["ov_load_1"]),
        ],
    )


def _create_lehe(net, buses):
    # Lehe
    buses["c_load_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe Industrielast/c_load_2",
        type="n",
        geodata=(53.571902, 8.596871),
    )
    pp.create_load(
        net,
        bus=buses["c_load_2"],
        p_mw=3.905,
        name="Lehe Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.905),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["ov_load_1"],
        to_bus=buses["c_load_2"],
        length_km=3.19,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["ov_load_1"]),
            _bus_geo(net, buses["c_load_2"]),
        ],
    )

    buses["c_lv_1_0"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Lehe LV - 0/c_lv_1_0",
        type="n",
        geodata=(53.574865, 8.598650),
    )

    pp.create_sgen(
        net,
        bus=buses["c_lv_1_0"],
        p_mw=0.41,
        name="PV Lehe - 0",
        min_p_mw=0,
        max_p_mw=0.41,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.41),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_1_0"],
        p_mw=3.3,
        name="Lehe Households - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.3),
    )

    buses["c_load_3_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe MV - 0/c_load_3_0",
        type="n",
        geodata=(53.560578, 8.593889),
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_3_0"],
        buses["c_lv_1_0"],
        sn_mva=2.2,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 1 - 0",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_2"],
        to_bus=buses["c_load_3_0"],
        length_km=1.27,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_2"]),
            _bus_geo(net, buses["c_load_3_0"]),
        ],
    )


def _create_windpark_speckenbuettel(net, buses):
    buses["ov_sgen_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Windpark Speckenbüttel 1/ov_sgen_1",
        type="n",
        geodata=(53.602913, 8.548237),
    )
    pp.create_sgen(
        net,
        bus=buses["ov_sgen_1"],
        p_mw=5.0,
        name="Windpark Speckenbüttel 1",
        type="WP",
        min_p_mw=0,
        max_p_mw=5.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=5),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_0"],
        to_bus=buses["ov_sgen_1"],
        length_km=0.905,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_0"]),
            _bus_geo(net, buses["ov_sgen_1"]),
        ],
    )


def _create_windpark_multibrid(net, buses):
    # # Multibrid M5000, 5 MW
    # # https://www.openstreetmap.org/node/1342407392
    buses["ov_sgen_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Windpark Multibrid/ov_sgen_2",
        type="n",
        geodata=(53.602913, 8.548237),
    )
    pp.create_sgen(
        net,
        bus=buses["ov_sgen_2"],
        p_mw=5.0,
        name="Windpark Multibrid",
        type="WP",
        min_p_mw=0,
        max_p_mw=5.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=5),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_0"],
        to_bus=buses["ov_sgen_2"],
        length_km=0.905,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_0"]),
            _bus_geo(net, buses["ov_sgen_2"]),
        ],
    )

    buses["c_load_12"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Geestemünde Industrielast/c_load_12",
        type="n",
        geodata=(53.530963, 8.599030),
    )
    pp.create_load(
        net,
        bus=buses["c_load_12"],
        p_mw=3.44,
        name="Geestemünde Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.44),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["ov_sgen_2"],
        to_bus=buses["c_load_12"],
        length_km=0.17,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["ov_sgen_2"]),
            _bus_geo(net, buses["c_load_12"]),
        ],
    )


def _create_geestemuende(net, buses):
    # Geestemünde
    buses["c_lv_3"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Geestemünde LV - 0/c_lv_3",
        type="n",
        geodata=(53.531437, 8.599357),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_3"],
        p_mw=3.45,
        name="Geestemünde Households - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.45),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_3"],
        p_mw=0.83,
        name="PV Geestemünde - 0",
        min_p_mw=0,
        max_p_mw=0.83,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.83),
    )

    buses["c_load_11"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Geestemünde MV - 0/c_load_11",
        type="n",
        geodata=(53.532464, 8.598488),
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_11"],
        buses["c_lv_3"],
        sn_mva=2.2,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 3",
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_12"],
        to_bus=buses["c_load_11"],
        length_km=0.389,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_12"]),
            _bus_geo(net, buses["c_load_11"]),
        ],
    )
    # Siemens SWT-2.3-93, 2.3 MW
    # https://www.openstreetmap.org/node/1145946171
    # Siemens SWT-2.3-93, 2.3 MW
    # https://www.openstreetmap.org/node/1145946189
    # Siemens SWT-2.3-93, 2.3 MW
    # https://www.openstreetmap.org/node/272116702
    # REpower 3.XM, 3.3 MW
    # https://www.openstreetmap.org/node/1342407395
    # Enercon E-82, 2 MW
    # https://www.openstreetmap.org/node/1305577597
    # Senvion 3.4M104, 3.4 MW
    # https://www.openstreetmap.org/node/4767958198
    # Multibrid M5000, 5 MW
    # https://www.openstreetmap.org/node/1230878049
    # Enercon Unknown, 3 MW
    # https://www.openstreetmap.org/node/5394329003
    # Enercon E-101, 3.05 MW
    # https://www.openstreetmap.org/node/4767958197
    # Enercon E-115, 3 MW
    # https://www.openstreetmap.org/node/617575631
    # Enercon E-82, 3 MW
    # https://www.openstreetmap.org/node/1342407391
    # REpower 3.3M, 3.3 MW
    # https://www.openstreetmap.org/node/1342407393


def _create_city(net, buses, use_extension):
    # city
    # ext grid right corner
    buses["c_ext_grid"] = pp.create_bus(
        net,
        vn_kv=20,
        name="110 kV Langen Bremerhaven Nord/c_ext_grid",
        type="b",
        geodata=(53.60537, 8.6241),
    )

    pp.create_ext_grid(net, bus=buses["c_hv_0"], name="BHV Nord")

    pp.create_transformer_from_parameters(
        net,
        buses["c_hv_0"],
        buses["c_ext_grid"],
        sn_mva=25,
        vn_hv_kv=110,
        vn_lv_kv=20,
        vkr_percent=0.16,
        vk_percent=12.00107,
        pfe_kw=0,
        i0_percent=0,
        shift_degree=30.0,
        name="Trafo 0-1",
    )

    # line between rings
    line_1 = pp.create_line_from_parameters(
        net,
        from_bus=buses["c_ext_grid"],
        to_bus=buses["c_load_11"],
        length_km=0.906,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_ext_grid"]),
            _bus_geo(net, buses["c_load_11"]),
        ],
    )
    # switch 0
    pp.create_switch(
        net, buses["c_load_11"], line_1, et="l", closed=True, type="LBS"
    )
    pp.create_switch(
        net,
        buses["c_ext_grid"],
        line_1,
        et="l",
        closed=False,
        type="LBS",
        name="S0",
    )
    buses["c_lv_1_1"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Lehe LV - 1/c_lv_1_1",
        type="n",
        geodata=(53.574865, 8.598650),
    )

    pp.create_sgen(
        net,
        bus=buses["c_lv_1_1"],
        p_mw=0.44,
        name="PV Lehe - 1",
        min_p_mw=0,
        max_p_mw=0.44,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.44),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_1_1"],
        p_mw=3.03,
        name="Lehe Households - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=3),
    )

    buses["c_load_3_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe MV - 1/c_load_3_1",
        type="n",
        geodata=(53.560578, 8.593889),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_ext_grid"],
        to_bus=buses["c_load_3_1"],
        length_km=1.27,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_ext_grid"]),
            _bus_geo(net, buses["c_load_3_1"]),
        ],
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_3_1"],
        buses["c_lv_1_1"],
        sn_mva=1.9,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 1 - 1",
    )
    buses["c_lv_1_2"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Lehe LV - 2/c_lv_1_2",
        type="n",
        geodata=(53.574865, 8.598650),
    )

    pp.create_sgen(
        net,
        bus=buses["c_lv_1_2"],
        p_mw=0.49,
        name="PV Lehe - 2",
        min_p_mw=0,
        max_p_mw=0.49,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.49),
    )

    pp.create_load(
        net,
        bus=buses["c_lv_1_2"],
        p_mw=3.3,
        name="Lehe Households - 2",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.3),
    )

    buses["c_load_3_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe MV - 2/c_load_3_2",
        type="n",
        geodata=(53.560578, 8.593889),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_1"],
        to_bus=buses["c_load_3_2"],
        length_km=1.27,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_1"]),
            _bus_geo(net, buses["c_load_3_2"]),
        ],
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_3_2"],
        buses["c_lv_1_2"],
        sn_mva=2.1,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 1 - 2",
    )
    if use_extension:
        _ex_create_pv_plant_for_city(net, buses)
    else:
        # ----- Erdgas -----
        # Kronos, 17,1 MW, 110/6 kV
        buses["c_sgen_0"] = pp.create_bus(
            net,
            vn_kv=20,
            name="110 kV Gen Bus Erdgas Kronos Titan GmbH/c_sgen_0",
            type="n",
            geodata=(53.528558, 8.547564),
        )
        pp.create_sgen(
            net,
            bus=buses["c_sgen_0"],
            p_mw=17.1,
            name="Erdgas Kronos Titan GmbH",
            type="GUD",
            min_p_mw=0,
            max_p_mw=17.1,
            q_mvar=compute_q(cos_phi=0.8, p_w=17.1),
        )
        # The gas and steam turbine power plant not only generates power,
        # it also draws some from the grid. Because we work with the
        # scaling and to avoid a negative scaling, we create a load at
        # the same bus with the values the gas and steam turbine power
        # plant draws from the grid while the sgen only considers the
        # generated power

        pp.create_load(
            net,
            bus=buses["c_sgen_0"],
            p_mw=0.04,
            name="Erdgas Kronos Titan GmbH",
            q_mvar=compute_q(cos_phi=0.97, p_w=0.04),
        )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_2"],
        to_bus=buses["c_sgen_0"],
        length_km=2.72,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_2"]),
            _bus_geo(net, buses["c_sgen_0"]),
        ],
    )
    # Geestemünde
    buses["c_lv_3"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Geestemünde LV - 1/c_lv_3",
        type="n",
        geodata=(53.531437, 8.599357),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_3"],
        p_mw=4.25,
        name="Geestemünde Households - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=4.25),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_3"],
        p_mw=0.83,
        name="PV Geestemünde - 1",
        min_p_mw=0,
        max_p_mw=0.83,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.83),
    )
    buses["c_load_11"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Geestemünde MV - 1/c_load_11",
        type="n",
        geodata=(53.532464, 8.598488),
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_11"],
        buses["c_lv_3"],
        sn_mva=2.9,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 3",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_sgen_0"],
        to_bus=buses["c_load_11"],
        length_km=0.389,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_sgen_0"]),
            _bus_geo(net, buses["c_load_11"]),
        ],
    )

    buses["c_lv_1_3"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Lehe LV - 3/c_lv_1_3",
        type="n",
        geodata=(53.574865, 8.598650),
    )

    pp.create_sgen(
        net,
        bus=buses["c_lv_1_3"],
        p_mw=0.45,
        name="PV Lehe - 3",
        min_p_mw=0,
        max_p_mw=0.45,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.45),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_1_3"],
        p_mw=3,
        name="Lehe Households - 3",
        q_mvar=compute_q(cos_phi=0.97, p_w=3),
    )

    buses["c_load_3_3"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe MV - 3/c_load_3_3",
        type="n",
        geodata=(53.560578, 8.593889),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_11"],
        to_bus=buses["c_load_3_3"],
        length_km=1.27,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_11"]),
            _bus_geo(net, buses["c_load_3_3"]),
        ],
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_3_3"],
        buses["c_lv_1_3"],
        sn_mva=2,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 1 - 3",
    )

    buses["c_lv_1_4"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Lehe LV - 4/c_lv_1_4",
        type="n",
        geodata=(53.574865, 8.598650),
    )

    pp.create_sgen(
        net,
        bus=buses["c_lv_1_4"],
        p_mw=0.44,
        name="PV Lehe - 4",
        min_p_mw=0,
        max_p_mw=0.44,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.44),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_1_4"],
        p_mw=3,
        name="Lehe Households - 4",
        q_mvar=compute_q(cos_phi=0.97, p_w=3),
    )

    buses["c_load_3_4"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Lehe MV - 4/c_load_3_4",
        type="n",
        geodata=(53.560578, 8.593889),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_3"],
        to_bus=buses["c_load_3_4"],
        length_km=1.27,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_3"]),
            _bus_geo(net, buses["c_load_3_4"]),
        ],
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_3_4"],
        buses["c_lv_1_4"],
        sn_mva=2,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 1",
    )
    # REpower 5M, 5 MW
    # https://www.openstreetmap.org/node/691722623
    buses["ov_sgen_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Windpark Repower/ov_sgen_2",
        type="n",
        geodata=(53.602913, 8.548237),
    )
    pp.create_sgen(
        net,
        bus=buses["ov_sgen_2"],
        p_mw=5.0,
        name="Windpark Repower",
        type="WP",
        min_p_mw=0,
        max_p_mw=5.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=5),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_4"],
        to_bus=buses["ov_sgen_2"],
        length_km=0.905,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_4"]),
            _bus_geo(net, buses["ov_sgen_2"]),
        ],
    )

    buses["c_lv_0"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Leherheide LV - 0/c_lv_0",
        type="n",
        geodata=(53.594306, 8.607556),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_0"],
        p_mw=3.7,
        name="Leherheide Households - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=3.7),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_0"],
        p_mw=0.46,
        name="PV Leherheide - 0",
        min_p_mw=0,
        max_p_mw=0.46,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.46),
    )
    buses["c_load_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Leherheide MV - 0/c_load_1",
        type="n",
        geodata=(53.592671, 8.604457),
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_1"],
        buses["c_lv_0"],
        sn_mva=2.4,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 0 - 0",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_3_4"],
        to_bus=buses["c_load_1"],
        length_km=0.3,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_3_4"]),
            _bus_geo(net, buses["c_load_1"]),
        ],
    )
    # ext grid central
    buses["c_ext_grid_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="110 kV Lehe ExtGrid/c_ext_grid_1",
        type="b",
        geodata=(53.563454, 8.601023),
    )

    pp.create_ext_grid(net, bus=buses["c_hv_1"], name="Lehe")

    pp.create_transformer(
        net,
        buses["c_hv_1"],
        buses["c_ext_grid_1"],
        name="110kV/20kV transformer",
        std_type="25 MVA 110/20 kV",
    )


def _ex_create_pv_plant_for_city(net, buses):
    buses["c_sgen_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="8.5MW wind plant for gas plant/c_sgen_0",
        type="n",
        geodata=(53.528558, 8.547564),
    )
    buses["c1_sgen_0"] = pp.create_bus(
        net, vn_kv=20, name="3.1MW pv plant for gas plant/c1_sgen_0", type="b"
    )
    buses["c2_sgen_0"] = pp.create_bus(
        net, vn_kv=20, name="1MW pv plant for gas plant/c2_sgen_0", type="b"
    )
    buses["c3_sgen_0"] = pp.create_bus(
        net, vn_kv=20, name="3MW wind plant for gas plant/c3_sgen_0", type="b"
    )
    buses["c4_sgen_0"] = pp.create_bus(
        net, vn_kv=20, name="1MW pv plant for gas plant/c4_sgen_0", type="b"
    )
    pp.create_sgen(
        net,
        bus=buses["c_sgen_0"],
        p_mw=8.5,
        name="8.5MW wind c_sgen_0",
        min_p_mw=0,
        max_p_mw=8.5,
        q_mvar=compute_q(cos_phi=0.9, p_w=8.5),
        type="wp",
    )  # 3 wind turbine ( 2.5MW, 3MW and 3MW )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_sgen_0"],
        to_bus=buses["c1_sgen_0"],
        length_km=3,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )

    pp.create_sgen(
        net,
        bus=buses["c1_sgen_0"],
        p_mw=3.1,
        name="3.1MW pv c1_sgen_0",
        min_p_mw=0,
        max_p_mw=3.1,
        q_mvar=compute_q(cos_phi=0.92, p_w=3.1),
        type="pv",
    )
    pp.create_load(
        net,
        bus=buses["c1_sgen_0"],
        p_mw=0.3,
        name="load c1_sgen_0",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.3),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c1_sgen_0"],
        to_bus=buses["c2_sgen_0"],
        length_km=5,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )
    pp.create_sgen(
        net,
        bus=buses["c2_sgen_0"],
        p_mw=1.0,
        name="1MW pv c2_sgen_0",
        min_p_mw=0,
        max_p_mw=1,
        q_mvar=compute_q(cos_phi=0.93, p_w=1),
        type="biomass",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c2_sgen_0"],
        to_bus=buses["c3_sgen_0"],
        length_km=2.7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )
    pp.create_sgen(
        net,
        bus=buses["c3_sgen_0"],
        p_mw=3.0,
        name="3MW wind c3_sgen_0",
        min_p_mw=0,
        max_p_mw=3,
        q_mvar=compute_q(cos_phi=0.95, p_w=3),
        type="wp",
    )
    pp.create_load(
        net,
        bus=buses["c3_sgen_0"],
        p_mw=0.7,
        name="load c3_sgen_0",
        q_mvar=compute_q(cos_phi=0.95, p_w=0.7),
    )


def _create_windpark_lehe(net, buses):
    # ----- Windenergie: Windpark Lehe, 10.0 MW HS/MS ------
    buses["c_sgen_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Windpark Lehe/c_sgen_2",
        type="n",
        geodata=(53.561192, 8.622811),
    )
    pp.create_sgen(
        net,
        bus=buses["c_sgen_2"],
        p_mw=10.0,
        name="Windpark Lehe",
        type="WP",
        min_p_mw=0,
        max_p_mw=10.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=10.0),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_ext_grid_1"],
        to_bus=buses["c_sgen_2"],
        length_km=1.462,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_ext_grid_1"]),
            _bus_geo(net, buses["c_sgen_2"]),
        ],
    )


def _create_klinikum_reinkenheide(net, buses):
    buses["c_load_20"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Klinikum Bremerhaven - Reinkenheide gGmbH/c_load_20",
        type="n",
        geodata=(53.537969, 8.636555),
    )

    pp.create_load(
        net,
        bus=buses["c_load_20"],
        p_mw=0.938,
        name="Klinikum Bremerhaven - Reinkenheide gGmbH",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.938),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_sgen_2"],
        to_bus=buses["c_load_20"],
        length_km=1.03,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.6,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_sgen_2"]),
            _bus_geo(net, buses["c_load_20"]),
        ],
    )


def _create_schiffdorferdamm(net, buses):
    # Schiffdorferdamm
    buses["c_lv_6"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Schiffdorferdamm LV/c_lv_6",
        type="n",
        geodata=(53.530717, 8.632589),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_6"],
        p_mw=0.13,
        name="PV Schiffdorferdamm",
        min_p_mw=0,
        max_p_mw=0.13,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.13),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_6"],
        p_mw=1.043,
        name="Schiffdorferdamm Households",
        q_mvar=compute_q(cos_phi=0.97, p_w=1.043),
    )
    buses["c_load_18"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Schiffdorferdamm MV/c_load_18",
        type="n",
        geodata=(53.530717, 8.632589),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_20"],
        to_bus=buses["c_load_18"],
        length_km=1.47,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_20"]),
            _bus_geo(net, buses["c_load_18"]),
        ],
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_18"],
        buses["c_lv_6"],
        sn_mva=0.7,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 6",
    )

    buses["c_load_19"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Schiffdorferdamm Industrielast/c_load_19",
        type="n",
        geodata=(53.529534, 8.630127),
    )
    pp.create_load(
        net,
        bus=buses["c_load_19"],
        p_mw=0.266,
        name="Schiffdorferdamm Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.266),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_18"],
        to_bus=buses["c_load_19"],
        length_km=0.209,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_18"]),
            _bus_geo(net, buses["c_load_19"]),
        ],
    )


def _create_eisarena(net, buses):
    # Eisarena
    buses["c_load_4"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Bremerhaven Eisarena, Stadthalle etc./c_load_4",
        type="n",
        geodata=(53.554570, 8.590115),
    )
    pp.create_load(
        net,
        bus=buses["c_load_4"],
        p_mw=0.54,
        name="Bremerhaven Eisarena, Stadthalle etc.",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.54),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_19"],
        to_bus=buses["c_load_4"],
        length_km=0.7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_19"]),
            _bus_geo(net, buses["c_load_4"]),
        ],
    )

    buses["c_load_7"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Mitte Industrielast/c_load_7",
        type="n",
        geodata=(53.542821, 8.581104),
    )
    pp.create_load(
        net,
        bus=buses["c_load_7"],
        p_mw=1.296,
        name="Mitte Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=1.296),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_4"],
        to_bus=buses["c_load_7"],
        length_km=1.44,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_4"]),
            _bus_geo(net, buses["c_load_7"]),
        ],
    )


def _create_mitte(net, buses):
    # Mitte
    buses["c_lv_2"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Mitte LV - 0/c_lv_2",
        type="n",
        geodata=(53.547459, 8.580126),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_2"],
        p_mw=2.55,
        name="Mitte Households - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=2.55),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_2"],
        p_mw=0.33,
        name="PV Mitte - 0",
        min_p_mw=0,
        max_p_mw=0.33,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.33),
    )
    buses["c_load_6"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Mitte MV - 0/c_load_6",
        type="n",
        geodata=(53.546510, 8.579083),
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_6"],
        buses["c_lv_2"],
        sn_mva=1.7,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 2",
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_7"],
        to_bus=buses["c_load_6"],
        length_km=0.35,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_7"]),
            _bus_geo(net, buses["c_load_6"]),
        ],
    )


def _create_leherheide(net, buses):
    # Leherheide
    buses["c_load_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Leherheide Industrielast/c_load_0",
        type="n",
        geodata=(53.598963, 8.629993),
    )

    pp.create_load(
        net,
        bus=buses["c_load_0"],
        p_mw=1.701,
        name="Leherheide Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=1.701),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_6"],
        to_bus=buses["c_load_0"],
        length_km=6.7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_6"]),
            _bus_geo(net, buses["c_load_0"]),
        ],
    )


def _create_wulsdorf(net, buses):
    # Wulsdorf
    buses["c_lv_4_1"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Wulsdorf LV - 0/c_lv_4_1",
        type="n",
        geodata=(53.503051, 8.605203),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_4_1"],
        p_mw=2.6,
        name="Wulsdorf Households - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=2.6),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_4_1"],
        p_mw=0.33,
        name="PV Wulsdorf - 0",
        min_p_mw=0,
        max_p_mw=0.33,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.33),
    )
    buses["c_load_13_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Wulsdorf MV - 0/c_load_13_1",
        type="n",
        geodata=(53.503051, 8.605203),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_0"],
        to_bus=buses["c_load_13_1"],
        length_km=10.7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_0"]),
            _bus_geo(net, buses["c_load_13_1"]),
        ],
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_13_1"],
        buses["c_lv_4_1"],
        sn_mva=1.8,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 4",
    )


def _create_abfall(net, buses, use_extension):
    # # ----- Abfall -----
    # # BEG, 14.0 MW, 20 kV
    # # https://www.openstreetmap.org/way/105301541
    # # https://www.wikidata.org/wiki/Q59315437
    buses["c_sgen_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="20 kV Gen Bus Abfall BEG mbh/c_sgen_1",
        type="n",
        geodata=(53.547946, 8.617964),
    )
    if use_extension:
        _ex_create_pv_plant_for_abfall(net, buses)
    else:
        pp.create_sgen(
            net,
            bus=buses["c_sgen_1"],
            p_mw=14.0,
            name="Abfall BEG mbh",
            min_p_mw=0,
            max_p_mw=14.0,
            q_mvar=compute_q(cos_phi=0.8, p_w=14),
            type="CHP",
        )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_13_1"],
        to_bus=buses["c_sgen_1"],
        length_km=5,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.55,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_13_1"]),
            _bus_geo(net, buses["c_sgen_1"]),
        ],
    )

    buses["c_load_17"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Surheide Industrielast/c_load_17",
        type="n",
        geodata=(53.517510, 8.632251),
    )

    pp.create_load(
        net,
        bus=buses["c_load_17"],
        p_mw=0.311,
        name="Surheide Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.311),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_sgen_1"],
        to_bus=buses["c_load_17"],
        length_km=3.5,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_sgen_1"]),
            _bus_geo(net, buses["c_load_17"]),
        ],
    )


def _ex_create_pv_plant_for_abfall(net, buses):
    # ( Replacement of Waste fuel powerplant by RE # 14MW )
    # C1_sgen_1 ,C2_sgen_1, C3_sgen_1, C4_sgen_1 buses are created
    # All these buses are connected to the existing bus C_sgen_1 in a
    # radial format
    # No transformer and LV line are connected yet in the system
    buses["c1_sgen_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="3MW wind plant for waste plant/c1_sgen_1",
        type="b",
    )
    buses["c2_sgen_1"] = pp.create_bus(
        net, vn_kv=20, name="2MW pv plant for waste plant/c2_sgen_1", type="b"
    )
    buses["c3_sgen_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="1.5MW pv plant for waste plant/c3_sgen_1",
        type="b",
    )
    buses["c4_sgen_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="2.5MW wind plant for waste plant/c4_sgen_1",
        type="b",
    )

    pp.create_sgen(
        net,
        bus=buses["c_sgen_1"],
        p_mw=5.0,
        name="5MW wind c_sgen_1",
        min_p_mw=0,
        max_p_mw=5.0,
        q_mvar=compute_q(cos_phi=0.9, p_w=5),
        type="wp",
    )  # 2 wind turbine, 2.5MW each

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_sgen_1"],
        to_bus=buses["c1_sgen_1"],
        length_km=1.28,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )
    pp.create_sgen(
        net,
        bus=buses["c1_sgen_1"],
        p_mw=3.0,
        name="3MW wind c1_sgen_1",
        min_p_mw=0,
        max_p_mw=3.0,
        q_mvar=compute_q(cos_phi=0.95, p_w=3),
        type="wp",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c1_sgen_1"],
        to_bus=buses["c2_sgen_1"],
        length_km=2,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )

    pp.create_sgen(
        net,
        bus=buses["c2_sgen_1"],
        p_mw=2.0,
        name="2MW solar c2_sgen_1",
        min_p_mw=0,
        max_p_mw=2.0,
        q_mvar=compute_q(cos_phi=0.97, p_w=2),
        type="pv",
    )
    pp.create_load(
        net,
        bus=buses["c2_sgen_1"],
        p_mw=0.7,
        name="load c2_sgen_1",
        q_mvar=compute_q(cos_phi=0.96, p_w=0.7),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c2_sgen_1"],
        to_bus=buses["c3_sgen_1"],
        length_km=7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )

    pp.create_sgen(
        net,
        bus=buses["c3_sgen_1"],
        p_mw=1.5,
        name="1.5MW solar c3_sgen_1",
        min_p_mw=0,
        max_p_mw=1.5,
        q_mvar=compute_q(cos_phi=0.95, p_w=1.5),
        type="pv",
    )

    pp.create_load(
        net,
        bus=buses["c3_sgen_1"],
        p_mw=0.63,
        name="load c3_sgen_1",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.63),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c3_sgen_1"],
        to_bus=buses["c4_sgen_1"],
        length_km=3,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )

    pp.create_sgen(
        net,
        bus=buses["c4_sgen_1"],
        p_mw=2.5,
        name="2.5MW wind c4_sgen_1",
        min_p_mw=0,
        max_p_mw=2.5,
        q_mvar=compute_q(cos_phi=0.96, p_w=2.5),
        type="biomass",
    )

    pp.create_load(
        net,
        bus=buses["c4_sgen_1"],
        p_mw=0.2,
        name="load1 c4_sgen_1",
        q_mvar=compute_q(cos_phi=0.95, p_w=0.2),
    )
    pp.create_load(
        net,
        bus=buses["c4_sgen_1"],
        p_mw=1.1,
        name="load2 c4_sgen_1",
        q_mvar=compute_q(cos_phi=0.93, p_w=1.1),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c4_sgen_1"],
        to_bus=buses["c_sgen_1"],
        length_km=5,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
    )


def _create_surheide(net, buses):
    # # Surheide
    buses["c_lv_5"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Surheide LV/c_lv_5",
        type="n",
        geodata=(53.519094, 8.637143),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_5"],
        p_mw=1.22,
        name="Surheide Households",
        q_mvar=compute_q(cos_phi=0.97, p_w=1.22),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_5"],
        p_mw=0.16,
        name="PV Surheide",
        min_p_mw=0,
        max_p_mw=0.16,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.16),
    )
    buses["c_load_16"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Surheide MV/c_load_16",
        type="n",
        geodata=(53.519094, 8.637143),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_17"],
        to_bus=buses["c_load_16"],
        length_km=0.37,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_17"]),
            _bus_geo(net, buses["c_load_16"]),
        ],
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_16"],
        buses["c_lv_5"],
        sn_mva=0.8,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 5",
    )


def _create_klinikum_buergerpark(net, buses):
    # Klinikum Bürgerpark
    buses["c_load_10"] = pp.create_bus(
        net,
        vn_kv=20,
        name="AMEOS Klinikum Am Bürgerpark Bremerhaven/c_load_10",
        type="n",
        geodata=(53.532376, 8.604369),
    )
    pp.create_load(
        net,
        bus=buses["c_load_10"],
        p_mw=0.28,
        name="AMEOS Klinikum Am Bürgerpark Bremerhaven",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.28),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_16"],
        to_bus=buses["c_load_10"],
        length_km=2.6,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_16"]),
            _bus_geo(net, buses["c_load_10"]),
        ],
    )


def _create_innenstadt(net, buses):
    # # Innenstadt
    buses["c_load_9"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Bremerhaven Innenstadt/c_load_9",
        type="n",
        geodata=(53.542787, 8.581111),
    )
    pp.create_load(
        net,
        bus=buses["c_load_9"],
        p_mw=0.355,
        name="Bremerhaven Innenstadt",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.355),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_10"],
        to_bus=buses["c_load_9"],
        length_km=1.9,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_10"]),
            _bus_geo(net, buses["c_load_9"]),
        ],
    )


def _create_zoo(net, buses):
    # Zoo
    buses["c_load_8"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Zoo/c_load_8",
        type="n",
        geodata=(53.544990, 8.570957),
    )
    pp.create_load(
        net,
        bus=buses["c_load_8"],
        p_mw=0.27,
        name="Zoo",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.27),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_9"],
        to_bus=buses["c_load_8"],
        length_km=0.713,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_9"]),
            _bus_geo(net, buses["c_load_8"]),
        ],
    )


def _create_klinikum_mitte(net, buses):
    # Klinikum Mitte
    buses["c_load_5"] = pp.create_bus(
        net,
        vn_kv=20,
        name="AMEOS Klinikum Mitte Bremerhaven/c_load_5",
        type="n",
        geodata=(53.552685, 8.575043),
    )
    pp.create_load(
        net,
        bus=buses["c_load_5"],
        p_mw=0.236,
        name="AMEOS Klinikum Mitte Bremerhaven",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.236),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_8"],
        to_bus=buses["c_load_5"],
        length_km=0.89,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_8"]),
            _bus_geo(net, buses["c_load_5"]),
        ],
    )

    buses["c_lv_2"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Mitte LV - 1/c_lv_2",
        type="n",
        geodata=(53.547459, 8.580126),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_2"],
        p_mw=2.55,
        name="Mitte Households - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=2.55),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_2"],
        p_mw=0.32,
        name="PV Mitte - 1",
        min_p_mw=0,
        max_p_mw=0.32,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.32),
    )
    buses["c_load_6"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Mitte MV/c_load_6",
        type="n",
        geodata=(53.546510, 8.579083),
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_6"],
        buses["c_lv_2"],
        sn_mva=1.8,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 2",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_5"],
        to_bus=buses["c_load_6"],
        length_km=0.74,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_5"]),
            _bus_geo(net, buses["c_load_6"]),
        ],
    )

    buses["c_lv_0_1"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Leherheide LV - 1/c_lv_0_1",
        type="n",
        geodata=(53.594306, 8.607559),
    )

    pp.create_load(
        net,
        bus=buses["c_lv_0_1"],
        p_mw=2.2,
        name="Leherheide Households - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=2.2),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_0_1"],
        p_mw=0.28,
        name="PV Leherheide - 1",
        min_p_mw=0,
        max_p_mw=0.28,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.28),
    )
    buses["c_load_1_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Leherheide MV - 1/c_load_1_1",
        type="n",
        geodata=(53.592671, 8.604461),
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_1_1"],
        buses["c_lv_0_1"],
        sn_mva=1.6,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 0 - 1",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_6"],
        to_bus=buses["c_load_1_1"],
        length_km=5.4,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_6"]),
            _bus_geo(net, buses["c_load_1_1"]),
        ],
    )

    buses["c_lv_4"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Wulsdorf LV - 1/c_lv_4",
        type="n",
        geodata=(53.503051, 8.605203),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_4"],
        p_mw=2,
        name="Wulsdorf Households - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=2),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_4"],
        p_mw=0.25,
        name="PV Wulsdorf - 1",
        min_p_mw=0,
        max_p_mw=0.25,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.25),
    )
    buses["c_load_13"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Wulsdorf MV - 1/c_load_13",
        type="n",
        geodata=(53.503051, 8.605203),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_1_1"],
        to_bus=buses["c_load_13"],
        length_km=9.13,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.5,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_1_1"]),
            _bus_geo(net, buses["c_load_13"]),
        ],
    )

    pp.create_transformer_from_parameters(
        net,
        buses["c_load_13"],
        buses["c_lv_4"],
        sn_mva=1.4,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 4",
    )

    # Multibrid M5000, 5 MW
    # https://www.openstreetmap.org/node/691695235
    # Multibrid M5000, 5 MW
    # https://www.openstreetmap.org/node/2432566699
    # Senvion 3.4M104 - 3.4 MW
    # https://www.openstreetmap.org/node/1476500908
    # Senvion 3.4M104 - 3.4 MW
    # https://www.openstreetmap.org/node/2917846569
    # Senvion 3.4M114 - 3.4 MW
    # https://www.openstreetmap.org/node/2509552694
    # REpower 3.0M122, 3 MW
    # https://www.openstreetmap.org/node/2097192036


def _create_fischereihafen(net, buses):
    # fishing port ring
    # line between rings
    buses["fp_load_6"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen Industrielast/fp_load_6",
        type="n",
        geodata=(53.509672, 8.584799),
    )
    line_0 = pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_6"],
        to_bus=buses["c_load_13"],
        length_km=1.54,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_6"]),
            _bus_geo(net, buses["c_load_13"]),
        ],
    )

    # switch 1
    pp.create_switch(
        net, buses["c_load_13"], line_0, et="l", closed=True, type="LBS"
    )
    pp.create_switch(
        net,
        buses["fp_load_6"],
        line_0,
        et="l",
        closed=False,
        type="LBS",
        name="S1",
    )

    pp.create_load(
        net,
        bus=buses["fp_load_6"],
        p_mw=0.02,
        name="Fischereihafen Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.02),
    )

    buses["c_load_14"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Wulsdorf Industrielast/c_load_14",
        type="n",
        geodata=(53.499859, 8.606736),
    )

    pp.create_load(
        net,
        bus=buses["c_load_14"],
        p_mw=1.176,
        name="Wulsdorf Industrielast",
        q_mvar=compute_q(cos_phi=0.97, p_w=1.176),
    )
    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_6"],
        to_bus=buses["c_load_14"],
        length_km=1.8,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_6"]),
            _bus_geo(net, buses["c_load_14"]),
        ],
    )


def _create_geestemuende_v2(net, buses):
    # Geestemünde
    buses["c_lv_3"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Geestemünde LV - 2/c_lv_3",
        type="n",
        geodata=(53.531437, 8.599357),
    )
    pp.create_load(
        net,
        bus=buses["c_lv_3"],
        p_mw=2.9,
        name="Geestemünde Households - 2",
        q_mvar=compute_q(cos_phi=0.97, p_w=2.9),
    )
    pp.create_sgen(
        net,
        bus=buses["c_lv_3"],
        p_mw=0.83,
        name="PV Geestemünde - 2",
        min_p_mw=0,
        max_p_mw=0.83,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.83),
    )
    buses["c_load_11"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Geestemünde MV - 2/c_load_11",
        type="n",
        geodata=(53.532464, 8.598488),
    )
    pp.create_transformer_from_parameters(
        net,
        buses["c_load_11"],
        buses["c_lv_3"],
        sn_mva=1.8,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo C - 3",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_14"],
        to_bus=buses["c_load_11"],
        length_km=3.69,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_14"]),
            _bus_geo(net, buses["c_load_11"]),
        ],
    )


def _create_bremerhaven_sued(net, buses):
    # Bremerhaven Süd
    buses["c_load_15"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Bremerhaven Süd/c_load_15",
        type="n",
        geodata=(53.489738, 8.602587),
    )
    pp.create_load(
        net,
        bus=buses["c_load_15"],
        p_mw=0.533,
        name="Bremerhaven Süd",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.533),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_11"],
        to_bus=buses["c_load_15"],
        length_km=5.7,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_11"]),
            _bus_geo(net, buses["c_load_15"]),
        ],
    )

    buses["fp_load_5"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen - 4/fp_load_5",
        type="n",
        geodata=(53.525959, 8.582308),
    )

    pp.create_load(
        net,
        bus=buses["fp_load_5"],
        p_mw=0.25,
        name="Fischereihafen - 4",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.25),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["c_load_15"],
        to_bus=buses["fp_load_5"],
        length_km=4.24,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["c_load_15"]),
            _bus_geo(net, buses["fp_load_5"]),
        ],
    )
    buses["fp_load_4"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen - 3/fp_load_4",
        type="n",
        geodata=(53.507744, 8.588815),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_5"],
        to_bus=buses["fp_load_4"],
        length_km=2.072,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_5"]),
            _bus_geo(net, buses["fp_load_4"]),
        ],
    )

    pp.create_load(
        net,
        bus=buses["fp_load_4"],
        p_mw=0.25,
        name="Fischereihafen - 3",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.25),
    )

    buses["fp_load_3"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen - 2/fp_load_3",
        type="n",
        geodata=(53.498069, 8.589991),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_4"],
        to_bus=buses["fp_load_3"],
        length_km=1.08,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_4"]),
            _bus_geo(net, buses["fp_load_3"]),
        ],
    )

    pp.create_load(
        net,
        bus=buses["fp_load_3"],
        p_mw=0.25,
        name="Fischereihafen - 2",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.25),
    )

    buses["fp_load_2"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen - 1/fp_load_2",
        type="n",
        geodata=(53.516097, 8.573791),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_3"],
        to_bus=buses["fp_load_2"],
        length_km=2.275,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_3"]),
            _bus_geo(net, buses["fp_load_2"]),
        ],
    )

    pp.create_load(
        net,
        bus=buses["fp_load_2"],
        p_mw=0.25,
        name="Fischereihafen - 1",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.25),
    )

    buses["fp_load_1"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen - 0/fp_load_1",
        type="n",
        geodata=(53.515457, 8.585792),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_2"],
        to_bus=buses["fp_load_1"],
        length_km=0.798,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_2"]),
            _bus_geo(net, buses["fp_load_1"]),
        ],
    )
    pp.create_load(
        net,
        bus=buses["fp_load_1"],
        p_mw=0.25,
        name="Fischereihafen - 0",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.25),
    )

    buses["fp_load_0"] = pp.create_bus(
        net,
        vn_kv=20,
        name="Fischereihafen MV/fp_load_0",
        type="n",
        geodata=(53.509672, 8.584799),
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_1"],
        to_bus=buses["fp_load_0"],
        length_km=0.647,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_1"]),
            _bus_geo(net, buses["fp_load_0"]),
        ],
    )
    buses["fp_lv__0"] = pp.create_bus(
        net,
        vn_kv=0.4,
        name="Fischereihafen LV/fp_lv__0",
        type="n",
        geodata=(53.509672, 8.584799),
    )
    pp.create_sgen(
        net,
        bus=buses["fp_lv__0"],
        p_mw=0.01,
        name="PV Fischereihafen",
        min_p_mw=0,
        max_p_mw=0.01,
        type="PV",
        q_mvar=compute_q(cos_phi=0.9, p_w=0.01),
    )
    pp.create_load(
        net,
        bus=buses["fp_lv__0"],
        p_mw=0.078,
        name="Fischereihafen Households",
        q_mvar=compute_q(cos_phi=0.97, p_w=0.078),
    )

    pp.create_transformer_from_parameters(
        net,
        buses["fp_load_0"],
        buses["fp_lv__0"],
        sn_mva=0.4,
        vn_hv_kv=20,
        vn_lv_kv=0.4,
        vkr_percent=1.325,
        vk_percent=4,
        pfe_kw=0.95,
        i0_percent=0.2375,
        tap_side="hv",
        tap_neutral=0,
        tap_min=-2,
        tap_max=2,
        tap_step_percent=2.5,
        tp_pos=0,
        shift_degree=150,
        name="MV-LV-Trafo FP",
    )

    buses["fp_ext_grid"] = pp.create_bus(
        net,
        vn_kv=20,
        name="110 kV Ext Grid Fischereihafen MV/fp_ext_grid",
        type="b",
        geodata=(53.515996, 8.606491),
    )

    pp.create_ext_grid(
        net, bus=buses["hv_bus_8"], name="Ext Grid Fischereihafen"
    )

    pp.create_transformer(
        net,
        buses["hv_bus_8"],
        buses["fp_ext_grid"],
        name="110kV/20kV transformer",
        std_type="25 MVA 110/20 kV",
    )

    pp.create_line_from_parameters(
        net,
        from_bus=buses["fp_load_0"],
        to_bus=buses["fp_ext_grid"],
        length_km=1.59,
        type="cs",
        r_ohm_per_km=0.01,
        x_ohm_per_km=0.05,
        max_i_ka=0.4,
        c_nf_per_km=10,
        geodata=[
            _bus_geo(net, buses["fp_load_0"]),
            _bus_geo(net, buses["fp_ext_grid"]),
        ],
    )


def _invert_geodata(net):
    for idx, row in net.bus_geodata.iterrows():
        row["x"], row["y"] = row["y"], row["x"]
        net.bus_geodata.loc[idx] = row
        for idx, row in net.line_geodata.iterrows():
            reversed_coords = []
            for coord in row["coords"]:
                reversed_coords.append(tuple(reversed(coord)))
            row["coords"] = reversed_coords
            net.line_geodata.loc[idx] = row


def _bus_geo(net, bus):
    x = net["bus_geodata"].loc[bus, "x"]
    y = net["bus_geodata"].loc[bus, "y"]
    return x, y


# grid = build_grid()

# pp.runpp(grid, numba=True)
# plt.simple_plot(grid, plot_line_switches=True)



import logging

from ..constraints.tap_changes import ConstraintTapChanges
from .base import GridElement

LOG = logging.getLogger(__name__)
MAX_PERCENT = 150.0


class PPTransformer(GridElement):
    @staticmethod
    def pp_key() -> str:
        return "trafo"

    @staticmethod
    def res_pp_key() -> str:
        return "res_trafo"

    def __init__(self, index, grid, value):
        super().__init__(index, grid, LOG)
        self.max_percent = MAX_PERCENT
        self.tap_pos = 0
        self.add_constraint(ConstraintTapChanges(self, value))

    def step(self, time):
        old_state = self.tap_pos
        self._check(time)

        self.set_value("tap_pos", self.tap_pos)
        return old_state != self.tap_pos



import unittest

import pandapower as pp

from midas_powergrid.custom.bhv import build_grid
from midas_powergrid.model.static import PandapowerGrid


class TestBremerhaven(unittest.TestCase):
    def test_powerflow(self):
        grid = build_grid()
        self.assertTrue(grid.res_bus.empty)
        try:
            pp.runpp(grid)
        except pp.LoadflowNotConverged:
            self.assertTrue(False, "Load flow failed")
        self.assertFalse(grid.res_bus.empty)

    def test_powerflow_with_extension(self):
        grid = build_grid(use_extension=True)
        self.assertTrue(grid.res_bus.empty)
        try:
            pp.runpp(grid)
        except pp.LoadflowNotConverged:
            self.assertTrue(False, "Load flow failed")
        self.assertFalse(grid.res_bus.empty)

    def test_with_constraints(self):
        gparams = {
            "gridfile": "bhv",
            "constraints": [
                ["line", 100],
                ["load", 0.02],
                ["sgen", 0.05],
                ["bus", 0.1],
            ],
            "disable_numba": False
        }

        grid = PandapowerGrid(gparams)

        try:
            grid.run_powerflow(0)
        except pp.LoadflowNotConverged:
            self.assertTrue(False, "Load flow failed")
        self.assertFalse(grid.grid._grid.res_bus.empty)


if __name__ == "__main__":
    unittest.main()
